@@ -1524,11 +1524,6 @@ class OperatorFilterFunctionInfidelity(CostFunction):
     """
     Calculates the infidelity with the filter function formalism.
 
-    ATTENTION: The filter function package sorts the noise operators in
-    lexicographic order! The order chosen in the `filter_function_h_n` must
-    correspond to the order in `filter_function_n_coeffs_deriv` and
-    the one in `noise_power_spec_density`.
-
     Parameters
     ----------
     solver: `Solver`
@@ -1543,9 +1538,8 @@ class OperatorFilterFunctionInfidelity(CostFunction):
         case, the same spectrum is taken for all noise operators, in the
         second, it is assumed that there are no correlations between different
         noise sources and thus there is one spectrum for each noise operator.
-        ATTENTION: The filter function package sorts the noise operators in
-        lexicographic order! The order chosen in the `filter_function_h_n` must
-        correspond to the order in `noise_power_spec_density`
+        The order of the noise terms must correspond to the order defined in
+        the solver by filter_function_h_n.
 
     omega: Sequence[float]
         The frequencies at which the integration is to be carried out.
@@ -1593,7 +1587,8 @@ class OperatorFilterFunctionInfidelity(CostFunction):
             pulse=self.solver.pulse_sequence,
             spectrum=self.noise_power_spec_density(self.omega),
             omega=self.omega,
-            cache_intermediates=True
+            cache_intermediates=True,
+            n_oper_identifiers=self.solver.filter_function_n_oper_identifiers,
         )
         return infidelity
 
@@ -1618,6 +1613,7 @@ class OperatorFilterFunctionInfidelity(CostFunction):
             spectrum=self.noise_power_spec_density(self.omega),
             omega=self.omega,
             control_identifiers=c_id,
+            n_oper_identifiers=self.solver.filter_function_n_oper_identifiers,
             n_coeffs_deriv=self.solver.filter_function_n_coeffs_deriv_vals
         )
         # what comes from ff:
