@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 # =============================================================================
-#     filter_functions
+#     qopt
 #     Copyright (C) 2020 Julian Teske, Forschungszentrum Juelich
 #
 #     This program is free software: you can redistribute it and/or modify
@@ -19,11 +19,11 @@
 #     Contact email: j.teske@fz-juelich.de
 # =============================================================================
 """
-This file contains a matrix class which implements the computationally
-expensive numeric calculations.
+This file contains an operator / matrix class which encapsulates the
+numeric operations.
 
-The matrices can be stored and handled either as dense or sparse
-matrices for the sake of encapsulation of the numeric methods.
+The operators can be stored as dense matrices and a sparse representation
+is planed.
 The most frequently used and computationally expensive function is
 the matrix exponential and its derivative. These operations are required to
 calculate the analytic solution of the Schroedinger and Lindblad master
@@ -71,8 +71,14 @@ import scipy.linalg as la
 
 from abc import ABC, abstractmethod
 from typing import Tuple, Optional, Union, Sequence
+from unittest import mock
+from warnings import warn
 
-from qutip import Qobj
+try:
+    from qutip import Qobj
+except ImportError:
+    warn('Qutip not installed. plot_bloch_vector_evolution() is not available')
+    Qobj = mock.Mock()
 
 
 VALID_SCALARS = [int, float, complex, np.int8, np.int16, np.int32, np.int64,
@@ -92,86 +98,8 @@ class OperatorMatrix(ABC):
 
     Attributes
     ----------
-    data:
+    data
         The stored data. Its type is defined in subclasses.
-
-    Methods
-    -------
-    dag:
-        Adjoint (dagger) of matrix. Also called hermitian conjugate.
-
-    tr:
-        Trace of matrix.
-
-    prop(tau):
-        The matrix exponential exp(A*tau) of the matrix A.
-
-    dexp(dirr, tau, compute_expm=False)
-        The deriative of the exponential in the given dirrection
-
-    clean:
-        Delete stored data.
-
-    copy:
-        Returns a deep copy of the object.
-
-    shape:
-        Returns the shape of the matrix.
-
-    conj(copy_):
-        Complex conjugate of the matrix. Returns a copy or operates in place.
-
-    conjugate(copy_):
-        Alias for conj.
-
-    transpose(copy_):
-        Transpose the matrix.
-
-    flatten:
-        Returns the flattened matrix.
-
-    prop(tau):
-        Convenience function to calculate the propagator of the Schroedinger
-        equation.
-
-    exp(tau, method):
-        Returns the matrix exponential.
-
-    dexp(direction, tau, method):
-        Returns the matrix exponential and its frechet derivative.
-
-    kron(other):
-        Kronecker matrix product.
-
-    identity_like: ControlMatrix
-        Returns an identity matrix of the same shape and type. Only for square
-        matrices!
-
-    spectral_decomposition(hermitian):
-        Eigenvalues and eigenvectors.
-
-    __getitem__(identifier):
-        Returns the matrix element corresponding to the identifier.
-
-    __add__(other):
-        Addition with other instances of the same type or instances of the
-        data storage type.
-
-    __iadd__(other):
-        In place addition with other instances of the same type or instances of
-        the data storage type.
-
-    __mul__(other):
-        Either Matrix multiplication or scalar multiplication.
-
-    __imul__(other):
-        Inplace matrix or scalar multiplication.
-
-    __rmul__(other):
-        Reflective scalar or matrix multiplication.
-
-TODO:
-    * implement element wise division for scalars
     """
 
     def __init__(self) -> None:
@@ -392,6 +320,14 @@ TODO:
         """
         pass
 
+    @abstractmethod
+    def __truediv__(self, other: 'OperatorMatrix') -> 'OperatorMatrix':
+        pass
+
+    @abstractmethod
+    def __itruediv__(self, other: 'OperatorMatrix') -> 'OperatorMatrix':
+        pass
+
     @property
     def shape(self) -> Tuple:
         """Returns the shape of the matrix. """
@@ -404,7 +340,7 @@ TODO:
         Parameters
         ----------
         index: tuple of int, length: 2
-            Index describing an entry in the marix.
+            Index describing an entry in the matrix.
 
         Returns
         -------
@@ -415,20 +351,35 @@ TODO:
         pass
 
     @abstractmethod
-    def dag(self, copy_: bool = True) -> Optional['OperatorMatrix']:
+    def __setitem__(self, key, value) -> None:
+        """ Sets the value at the position key.
+
+        Parameters
+        ----------
+        key: tuple of int, length: 2
+            Index specifying an entry in the matrix.
+
+        value: complex
+            Value to be set at the position key.
+
+        """
+        pass
+
+    @abstractmethod
+    def dag(self, do_copy: bool = True) -> Optional['OperatorMatrix']:
         """
         Adjoint (dagger) of the matrix.
 
         Parameters
         ----------
-        copy_: bool, optional
+        do_copy: bool, optional
             If false, then the operation is executed inplace. Otherwise returns
             a new instance. Defaults to True.
 
         Returns
         -------
         out: OperatorMatrix
-            If copy_ is true, then a new instance otherwise self.
+            If do_copy is true, then a new instance otherwise self.
 
         """
         return self
@@ -446,42 +397,74 @@ TODO:
         return 0j
 
     @abstractmethod
-    def conj(self, copy_: bool = True) -> Optional['OperatorMatrix']:
+    def ptrace(self,
+               dims: Sequence[int],
+               remove: Sequence[int],
+               do_copy: bool = True) -> 'OperatorMatrix':
         """
-        Complex conjugate of the matrix.
+        Partial trace of the matrix.
 
+        If the matrix describes a ket, the corresponding density matrix is
+        calculated and used for the partial trace.
         Parameters
         ----------
-        copy_: bool, optional
-            If false, then the operation is executed inplace. Otherwise returns
+        dims : list of int
+            Dimensions of the subspaces making up the total space on which
+            the matrix operates. The product of elements in 'dims' must be
+            equal to the matrix' dimension.
+        remove : list of int
+            The selected subspaces over which the partial trace is formed.
+            The given indices correspond to the ordering of subspaces that
+            are specified via the 'dim' argument.
+        do_copy : bool, optional
+            If false, the operation is executed inplace. Otherwise returns
             a new instance. Defaults to True.
 
         Returns
         -------
-        out: OperatorMatrix
-            If copy_ is true, then a new instance otherwise self.
+        pmat : OperatorMatrix
+            The partially traced OperatorMatrix.
 
         """
         pass
 
-    def conjugate(self, copy_: bool = True) -> Optional['OperatorMatrix']:
-        """Alias for conj. """
-        return self.conj(copy_=copy_)
-
     @abstractmethod
-    def transpose(self, copy_: bool = True) -> Optional['OperatorMatrix']:
-        """Transpose of the matrix.
+    def conj(self, do_copy: bool = True) -> Optional['OperatorMatrix']:
+        r"""
+        Complex conjugate of the matrix.
 
         Parameters
         ----------
-        copy_: bool, optional
+        do_copy : bool, optional
             If false, then the operation is executed inplace. Otherwise returns
             a new instance. Defaults to True.
 
         Returns
         -------
         out: OperatorMatrix
-            If copy_ is true, then a new instance otherwise self.
+            If do_copy is true, then a new instance otherwise self.
+
+        """
+        pass
+
+    def conjugate(self, do_copy: bool = True) -> Optional['OperatorMatrix']:
+        """Alias for conj. """
+        return self.conj(do_copy=do_copy)
+
+    @abstractmethod
+    def transpose(self, do_copy: bool = True) -> Optional['OperatorMatrix']:
+        """Transpose of the matrix.
+
+        Parameters
+        ----------
+        do_copy: bool, optional
+            If false, then the operation is executed inplace. Otherwise returns
+            a new instance. Defaults to True.
+
+        Returns
+        -------
+        out: OperatorMatrix
+            If do_copy is true, then a new instance otherwise self.
 
         """
 
@@ -662,6 +645,36 @@ TODO:
         """
         pass
 
+    @classmethod
+    def pauli_0(cls):
+        """Pauli 0 i.e. the Identity matrix. """
+        return cls(np.eye(2))
+
+    @classmethod
+    def pauli_x(cls):
+        """Pauli x Matrix. """
+        return cls(np.asarray([[0, 1], [1, 0]]))
+
+    @classmethod
+    def pauli_y(cls):
+        """Pauli y Matrix. """
+        return cls(np.asarray([[0, -1j], [1j, 0]]))
+
+    @classmethod
+    def pauli_z(cls):
+        """Pauli z Matrix. """
+        return cls(np.diag([1, -1]))
+
+    @classmethod
+    def pauli_m(cls):
+        """Pauli minus Matrix i.e. descending operator. """
+        return cls(np.asarray([[0, 0], [1, 0]]))
+
+    @classmethod
+    def pauli_p(cls):
+        """Pauli plus Matrix i.e. ascending operator. """
+        return cls(np.asarray([[0, 1], [0, 0]]))
+
 
 class DenseOperator(OperatorMatrix):
     """
@@ -776,13 +789,32 @@ class DenseOperator(OperatorMatrix):
             raise NotImplementedError(str(type(other)))
         return self
 
+    def __truediv__(self, other: 'DenseOperator') -> 'DenseOperator':
+        if isinstance(other, (np.ndarray, *VALID_SCALARS)):
+            return DenseOperator(self.data / other)
+        raise NotImplementedError(str(type(other)))
+
+    def __itruediv__(self, other: 'DenseOperator') -> 'DenseOperator':
+        if isinstance(other, (np.ndarray, *VALID_SCALARS)):
+            self.data /= other
+            return self
+        raise NotImplementedError(str(type(other)))
+
     def __getitem__(self, index: Tuple) -> np.complex128:
         """See base class. """
         return self.data[index]
 
-    def dag(self, copy_: bool = True) -> Optional['DenseOperator']:
+    def __setitem__(self, key, value) -> None:
         """See base class. """
-        if copy_:
+        self.data.__setitem__(key, value)
+
+    def __repr__(self):
+        """Representation as numpy array. """
+        return 'DenseOperator with data: \n' + self.data.__repr__()
+
+    def dag(self, do_copy: bool = True) -> Optional['DenseOperator']:
+        """See base class. """
+        if do_copy:
             cp = self.copy()
             np.conj(cp.data, out=cp.data)
             cp.data = np.copy(cp.data.T)
@@ -792,9 +824,9 @@ class DenseOperator(OperatorMatrix):
             self.data = self.data.T
             return self
 
-    def conj(self, copy_: bool = True) -> Optional['DenseOperator']:
+    def conj(self, do_copy: bool = True) -> Optional['DenseOperator']:
         """See base class. """
-        if copy_:
+        if do_copy:
             copy = self.copy()
             np.conj(copy.data, out=copy.data)
             return copy
@@ -802,9 +834,9 @@ class DenseOperator(OperatorMatrix):
             np.conj(self.data, out=self.data)
             return self
 
-    def transpose(self, copy_: bool = True) -> Optional['DenseOperator']:
+    def transpose(self, do_copy: bool = True) -> Optional['DenseOperator']:
         """See base class. """
-        if copy_:
+        if do_copy:
             out = self.copy()
         else:
             out = self
@@ -839,6 +871,93 @@ class DenseOperator(OperatorMatrix):
         """See base class. """
         return self.data.trace()
 
+    def ptrace(self,
+               dims: Sequence[int],
+               remove: Sequence[int],
+               do_copy: bool = True) -> 'DenseOperator':
+        """
+        Partial trace of the matrix.
+
+        If the matrix describes a ket, the corresponding density matrix is
+        calculated and used for the partial trace.
+
+        This implementation closely follows that of QuTip's qobj._ptrace_dense.
+        Parameters
+        ----------
+        dims : list of int
+            Dimensions of the subspaces making up the total space on which
+            the matrix operates. The product of elements in 'dims' must be
+            equal to the matrix' dimension.
+        remove : list of int
+            The selected subspaces as indices over which the partial trace is
+            formed. The given indices correspond to the ordering of
+            subspaces specified in the 'dim' argument.
+        do_copy : bool, optional
+            If false, the operation is executed inplace. Otherwise returns
+            a new instance. Defaults to True.
+
+        Returns
+        -------
+        pmat : OperatorMatrix
+            The partially traced OperatorMatrix.
+
+        Raises
+        ------
+        AssertionError:
+            If matrix dimension does not match specified dimensions.
+
+        Examples
+        --------
+         ghz_ket = DenseOperator(np.array([[1,0,0,0,0,0,0,1]]).T) / np.sqrt(2)
+         ghz_rho = ghz_ket * ghz_ket.dag()
+         ghz_rho.ptrace(dims=[2,2,2], remove=[0,2])
+        DenseOperator with data:
+        array([[0.5+0.j, 0. +0.j],
+               [0. +0.j, 0.5+0.j]])
+        """
+
+        if self.shape[1] == 1:
+            mat = (self * self.dag()).data
+        else:
+            mat = self.data
+        if mat.shape[0] != np.prod(dims):
+            raise AssertionError("Specified dimensions do not match "
+                                 "matrix dimension.")
+        n_dim = len(dims)  # number of subspaces
+        dims = np.asarray(dims, dtype=int)
+
+        remove = list(np.sort(remove))
+        # indices of subspace that are kept
+        keep = list(set(np.arange(n_dim)) - set(remove))
+
+        dims_rm = (dims[remove]).tolist()
+        dims_keep = (dims[keep]).tolist()
+        dims = list(dims)
+
+        # 1. Reshape: Split matrix into subspaces
+        # 2. Transpose: Change subspace/index ordering such that the subspaces
+        # over which is traced correspond to the first axes
+        # 3. Reshape: Merge each, subspaces to be removed (A) and to be kept
+        # (B), common spaces/axes.
+        # The trace of the merged spaces (A \otimes B) can then be
+        # calculated as Tr_A(mat) using np.trace for input with
+        # more than two axes effectively resulting in
+        # pmat[j,k] = Sum_i mat[i,i,j,k] for all j,k = 0..prod(dims_keep)
+        pmat = np.trace(mat.reshape(dims + dims)
+                           .transpose(remove + [n_dim + q for q in remove] +
+                                      keep + [n_dim + q for q in keep])
+                           .reshape([np.prod(dims_rm),
+                                    np.prod(dims_rm),
+                                    np.prod(dims_keep),
+                                    np.prod(dims_keep)])
+                        )
+
+        if do_copy:
+            return DenseOperator(pmat)
+        else:
+            self.data = pmat
+            return self
+
     def kron(self, other: 'DenseOperator') -> 'DenseOperator':
         """See base class. """
         if type(other) == DenseOperator:
@@ -871,10 +990,10 @@ class DenseOperator(OperatorMatrix):
 
         """
         if is_skew_hermitian:
-            eig_val, eig_vec = la.eigh(-1j * self.data)
+            eig_val, eig_vec = np.linalg.eigh(-1j * self.data)
             eig_val = 1j * eig_val
         else:
-            eig_val, eig_vec = la.eig(self.data)
+            eig_val, eig_vec = np.linalg.eig(self.data)
 
         # apply the exponential function to the eigenvalues and invert the
         # diagonalization transformation
@@ -917,10 +1036,10 @@ class DenseOperator(OperatorMatrix):
 
         """
         if is_skew_hermitian:
-            eig_val, eig_vec = la.eigh(-1j * self.data)
+            eig_val, eig_vec = np.linalg.eigh(-1j * self.data)
             eig_val = 1j * eig_val
         else:
-            eig_val, eig_vec = la.eig(self.data)
+            eig_val, eig_vec = np.linalg.eig(self.data)
 
         eig_vec_dag = eig_vec.conj().T
 
@@ -1116,7 +1235,7 @@ class DenseOperator(OperatorMatrix):
             d_m = (self.data + epsilon * direction.data) * tau
             dprop = la.expm(d_m)
             prop = self.exp(tau)
-            prop_grad = (dprop - prop) * (1 / epsilon)
+            prop_grad = (DenseOperator(dprop) - prop) * (1 / epsilon)
 
         elif method == "first_order":
             if compute_expm:
@@ -1194,13 +1313,13 @@ class SparseOperator(OperatorMatrix):
 
 def convert_unitary_to_super_operator(
         unitary: Union['OperatorMatrix', np.array]):
-    """
+    r"""
     We assume that the unitary U shall be used to propagate a density matrix m
     like
 
     .. math::
 
-    U m U^dag
+        U m U^dag
 
     which is equivalent to
 
@@ -1216,7 +1335,7 @@ def convert_unitary_to_super_operator(
     Returns
     -------
     unitary_super_operator:
-        The unitary propagator in the Lindblad formalism.
+        The unitary propagator in the Lindblad formalism. Same type as input.
 
     Raises
     ------
@@ -1225,7 +1344,7 @@ def convert_unitary_to_super_operator(
 
     """
     if type(unitary) in (DenseOperator, SparseOperator):
-        return unitary.conj(copy_=True).kron(unitary)
+        return unitary.conj(do_copy=True).kron(unitary)
     elif isinstance(unitary, np.ndarray):
         return np.kron(np.conj(unitary), unitary)
     else:
@@ -1233,6 +1352,86 @@ def convert_unitary_to_super_operator(
                          'numpy array!')
 
 
+def ket_vectorize_density_matrix(
+        density_matrix: Union['OperatorMatrix', np.array]):
+    r"""
+    Vectorizes a density matrix column-wise as ket vector.
+
+    Parameters
+    ----------
+    density_matrix: OperatorMatrix or numpy array
+        The density matrix.
+
+    Returns
+    -------
+    density_ket_vector:
+        The density matrix as ket vector for the Liouville formalism.
+
+    Raises
+    ------
+    ValueError:
+        If the operation is not defined for the input type.
+
+    ValueError:
+        If the density matrix is not given in square shape.
+
+    """
+    if not density_matrix.shape[0] == density_matrix.shape[1]:
+        raise ValueError('The density matrix must be of square shape. ')
+    if type(density_matrix) in (DenseOperator, SparseOperator):
+        vectorized_matrix = density_matrix.copy()
+        vectorized_matrix.data = vectorized_matrix.data.T.reshape(
+            [vectorized_matrix.data.size, 1]
+        )
+        return vectorized_matrix
+    elif isinstance(density_matrix, np.ndarray):
+        return density_matrix.T.reshape([density_matrix.size, 1])
+    else:
+        raise ValueError('The target must be given as dense control matrix or '
+                         'numpy array!')
+
+
+def convert_ket_vectorized_density_matrix_to_square(
+        vectorized_density_matrix: Union['OperatorMatrix', np.array]):
+    r"""
+    Bring vectorized density matrix back into square form.
+
+    Parameters
+    ----------
+    vectorized_density_matrix: OperatorMatrix or numpy array
+        The density matrix.
+
+    Returns
+    -------
+    density_ket_vector:
+        The density matrix as ket vector for the Liouville formalism.
+
+    Raises
+    ------
+    ValueError:
+        If the operation is not defined for the input type.
+
+    ValueError:
+        If the density matrix is not given as ket vector.
+
+    """
+    if not vectorized_density_matrix.shape[1] == 1:
+        raise ValueError('The density matrix must be vectorized as ket. ')
+    if type(vectorized_density_matrix) in (DenseOperator, SparseOperator):
+        d = int(np.sqrt(vectorized_density_matrix.data.size))
+        vectorized_matrix = vectorized_density_matrix.copy()
+        vectorized_matrix.data = vectorized_matrix.data.reshape(
+            [d, d]
+        ).T
+        return vectorized_matrix
+    elif isinstance(vectorized_density_matrix, np.ndarray):
+        d = int(np.sqrt(vectorized_density_matrix.size))
+        return vectorized_density_matrix.reshape([d, d]).T
+    else:
+        raise ValueError('The target must be given as dense control matrix or '
+                         'numpy array!')
+
+
 def closest_unitary(matrix: OperatorMatrix):
     """
     Calculate the unitary matrix U that is closest with respect to the
